/*
 *      C r e a t e . c
 *
 *  Functions used in creating lists.
 *
 *
 * (c) Copyright 1996-1997 David J. Walker
 *     Email:  d.j.walker@x400.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  18 May 96   DJW   - First version
 *
 *  14 July 96  DJW   - Merged in 'Clone' variants to this file
 */

#include <liblist.h>
#include "liblistx.h"
#include <stdlib.h>
#include <string.h>

/*======================================================== LIST_CREATE */
extern
list_t  LIST_Create  (struct _listclass * listclass, 
/*      ~~~~~~~~~~~*/ size_t nodesize,
                      int (*initfunc)(node_t,va_list), 
                      int (*killfunc)(node_t),
                      int (*compfunc)(node_t, node_t) )
/*
 *  This is the main create routine used to
 *  instantiate new lists.
 *---------------------------------------------------------------------*/
{
    listclass_t newclass;
    list_t      newlist;

    LIBDEBUG ("LIST_Create", \
           "type parameter is NULL", \
           listclass != NULL)

    if (nodesize == 0 && listclass->master->listtype != LIST_TYPE_EMBEDDED)
    {
        errno = LIST_ERROR_BADSIZE;
        return NULL;
    }

	/*
	 *  Now do some checks that are specific
	 *  to the list type requested.
	 */
    switch (listclass->master->listtype)
    {
    case LIST_TYPE_FIFO:
    case LIST_TYPE_LIFO:
            /*
			 *  Certain table types by their very nature
			 *  are always unordered. We must NOT have a
			 *  compare function specified.
			 */
		    if (compfunc != NULL)
            {
                errno = LIST_ERROR_BADCOMP;
                return NULL;
            }
            break;
    default:
            break;
    }
    /*
     *  Set up this list
     */
    newlist = _LIST_CloneClass(listclass);
    if (newlist == NULL)
    {
        return NULL;
    }
	newclass = newlist->listclass;
    newclass->sizeuser = nodesize;
    newclass->initnode = initfunc;
    newclass->killnode = killfunc;
    newclass->compare = compfunc;
    return newlist;
}


/*========================================================== LIST_CLONE */
extern
list_t  LIST_Clone  (list_t oldlist)
/*		~~~~~~~~~~
 *	Public entry point for cloning purposes.
 *
 *	Merely does any special validation of parameters,
 *	and then calls list type specific routine.
 *----------------------------------------------------------------------*/
{
    LIBDEBUG ("LIST_Clone", \
           "'oldlist' does not point to a valid list type", \
           _LIST_ListCheck(oldlist))

    return (oldlist->listclass->master->clone)(oldlist);
}


/*==================================================== _LIST_CLONELIST */
extern
list_t  _LIST_CloneList  (list_t oldlist)
/*		~~~~~~~~~~~~~~~
 *  Internal support routine that does the
 *  clone process given the list pointer.
 *  This does generic work required via most
 *  the different list formats.  Some of the
 *  formats return direct to the user, and
 *  others return to the format specific clone
 *  routine to do further processing.
 *---------------------------------------------------------------------*/
{
    /*
     *  Set up a new list instance, and
     *  ensure that all fields are clear
     */
	LIBASSERT(_LIST_ListCheck(oldlist))
	return _LIST_CloneClass(oldlist->listclass);
}


/*=================================================== _LIST_CLONECLASS */
extern
list_t  _LIST_CloneClass  (listclass_t listclass)
/*		~~~~~~~~~~~~~~~~
 *  Internal support routine that does the
 *  clone process given the class pointer.
 *---------------------------------------------------------------------*/
{
	listclass_t newclass;
	list_t		newlist;

	/*
	 *	The first think to do is to decide if we
	 *	need to create a new 'listclass' area.
	 */
	if (listclass->sizeuser != 0)
	{
		/*
		 *	This is not a master instance,
		 *	so we can use it.
		 */
		LIBASSERT (listclass->clones != 0)
		newclass = listclass;
	}
	else
	{
	    /*
		 *  Create a new class structure
	     *  from a master instance.
		 */
		LIBASSERT (listclass->clones == 0)
	    newclass = malloc (sizeof(struct _listclass));
		if (newclass == NULL)
	    {
			errno = LIST_ERROR_NOMEMORY;
	        return NULL;
		}
		(void) memcpy (newclass, listclass, sizeof (struct _listclass));
	}

	/*
	 *	Now it is time to create the
	 *	actual list instance we want.
	 */
    LIBASSERT (newclass->sizelist > 0)
    newlist = calloc (newclass->sizelist,1);
    if (newlist == NULL)
    {
		if (newclass->sizeuser == 0)
		{
			free (newclass);
		}
		errno = LIST_ERROR_NOMEMORY;
        return NULL;
    }

    newlist->listclass = newclass;
    /*
     *  Increment clone count
     */
    newclass->clones += 1;
    /*
     *  And return to caller
     */
#ifdef LISTTRACK
    (void) _LIST_TrackAdd (newlist);
#endif /* LISTTRACK */
    return newlist;
}


/*========================================================= _LIST_UNCLONE */
extern
void	_LIST_UnClone (list_t list)
/*		~~~~~~~~~~~~~
 *  Internal support routine that is used  
 *  when we have freed all the nodes, so we
 *  can destroy the list object itself.
 *------------------------------------------------------------------------*/
{
	listclass_t	listclass = list->listclass;
#ifdef DEBUG
    (void)memset (list, 0, listclass->sizelist);
#endif /* DEBUG */
    free (list);
    /*
     *  If this is the last clone,
     *  then free other resources as well
     */
    listclass->clones -= 1;
    if (listclass->clones == 0)
    {
#ifdef DEBUG
        (void)memset (listclass, 0, sizeof(struct _listclass));
#endif /* DEBUG */
        free (listclass);
    }
	return;
}
