/*
 *      h a s h s h a r _ c
 *
 *  This contains routines that are specific
 *  to the implementation of hash tables, but
 *  that contain no code that would stop their
 *  inclusion into the RLL library variant.
 *
 *
 * (c) Copyright 1996-1997 David J. Walker
 *     Email:  d.j.walker@x400.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  24 May 96   DJW   - First version
 *
 *	28 Mar 98	DJW	  -	Fixed problem in _LIST_HashFind() routine when a
 *						bucket has not yet had a list attached to it.
 *						(problem and suggested fix from Gerald Stegmiller)
 */

#include <liblist.h>
#include "liblistx.h"


/*========================================================= LIST_HASHBUCKET */
extern
_hash_t	_LIST_HashBucket (_hashh_t hashlist, node_t node)
/*
 *	Call the correct hashing algorithm, normalise
 *	the return value to fit into the table, and
 *	calculate the address of the bucket concerned.
 *------------------------------------------------------------------------*/
{
	LIBASSERT(hashlist->bucketcount != 0)

    return &hashlist->buckets[hashlist->nodehash(_LIST_USERNODE(hashlist,node))
							  % hashlist->bucketcount];
}


/*========================================================== HASH_CHECKBUCKET */
static
int		HASH_CheckBucket (_hashh_t hashlist, _hash_t  bucket)
/*		~~~~~~~~~~~~~~~~
 *	Check that there is a list currently set up
 *	and attached to this node.  If not try and
 *	set one up.   Report success or failure.
 *---------------------------------------------------------------------------*/
{
	if (bucket->u.list == NULL)
	{
		bucket->u.list = hashlist->clonelist->listclass->master->clone(hashlist->clonelist);
		if (bucket->u.list == NULL)
		{
			return LIST_ERROR_NOMEMORY;
		}
	}
	return LIST_ERROR_NONE;
}


/*========================================================= HASH_FIRST */
extern
node_t    _LIST_HashFirst  (list_t list)
/*        ~~~~~~~~~~~~~~~
 *  This function will find the first node by
 *  starting at the base and following left
 *  branches until a leaf is reached.
 *----------------------------------------------------------------------*/
{
	_hashh_t   hashlist = (_hashh_t)list;
    _hash_t    bucket, endptr;
    node_t     node;

	/*
	 *	Check if list set up yet.
	 */
	if (hashlist->buckets == NULL)
	{
		errno = LIST_ERROR_NOTSETUP;
		return NULL;
	}

    for (bucket = &hashlist->buckets[0], endptr = &hashlist->buckets[hashlist->bucketcount];
                bucket < endptr ; bucket++)
    {
        /*
         *  If there is no linked table,
         *  then the bucket contains the
         *  node pointer.
         */
        if (hashlist->clonelist == NULL)
        {
			if (bucket->u.node == NULL)
			{
				continue;
			}
            return _LIST_USERNODE(hashlist,bucket);
        }
        /*
         *  If there is a linked list, then
         *  check it out to see if it has data.
         */
		if (bucket->u.list == NULL)
		{
			continue;
		}
        node = bucket->u.list->listclass->master->first(bucket->u.list);
        if (node != NULL)
        {
            return (node);
        }
    }
    return (NULL);
}


/*========================================================== HASH_LAST */
extern
node_t    _LIST_HashLast  (list_t list)
/*        ~~~~~~~~~~~~~~
 *  This function will find the last node by
 *  working backwards through the buckets.
 *  If there is a linked table (the norm) then
 *  it will call the LIST_Last method on the
 *  linked table.  It will repeat this process
 *  until either it runs out of buckets, or one
 *  of the calls returns a valid node.
 *----------------------------------------------------------------------*/
{
	_hashh_t   hashlist = (_hashh_t)list;
    _hash_t    bucket;
    node_t     node;

	/*
	 *	Check if list set up yet.
	 */
	if (hashlist->buckets == NULL)
	{
		errno = LIST_ERROR_NOTSETUP;
		return NULL;
	}
    for (bucket = &hashlist->buckets[hashlist->bucketcount - 1]; 
                            bucket >= &hashlist->buckets[0] ; bucket--)
    {
        /*
         *  If there is no linked table,
         *  then the bucket contains the
         *  node pointer.
         */
        if (hashlist->clonelist == NULL)
        {
			if (bucket->u.node == NULL)
			{
				continue;
			}
            return _LIST_USERNODE(hashlist,bucket->u.node);
        }
        /*
         *  If there is a linked list, then
         *  check it out to see if it has data.
         */
		if (bucket->u.list == NULL)
		{
			continue;
		}
        node = bucket->u.list->listclass->master->last(bucket->u.list);
        if (node != NULL)
        {
            return (node);
        }
    }
    return (NULL);
}


/*======================================================== HASH_PREVIOUS */
extern
node_t  _LIST_HashPrevious      (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~~~~~
 *  To find the previous node, we start at the bucket containing
 *	the current node.  We search for previous nodes in attached
 *	to that bucket.  If not we move to the previous bucket and
 *	repeat the process.  This carries on until we either
 *	find another node or reach the start of the hash table.
 *-----------------------------------------------------------------------*/
{
	_hashh_t hashlist = (_hashh_t)list;
    _hash_t  bucket;
    _hash_t  firstbucket;
    node_t   nextnode;

	/*
	 *	Check if list set up yet.
	 */
	if (hashlist->buckets == NULL)
	{
		errno = LIST_ERROR_NOTSETUP;
		return NULL;
	}

    for (bucket = _LIST_HashBucket(hashlist,node), nextnode = node, 
	     firstbucket = &hashlist->buckets[0] ; 
				bucket >= firstbucket; 
						bucket--)
    {
		if (hashlist->clonelist == NULL)
		{
			if (bucket->u.node == NULL)
			{
				continue;
			}
			if (bucket->u.node != node)
			{
				return bucket->u.node;
			}
		}

		/*
		 *	If list empty, no point in looking
		 */
		if (bucket->u.list == NULL)
		{
			continue;
		}
		/*
		 *	Call last/previous as appropriate
		 */
		nextnode = (nextnode == NULL)
						? bucket->u.list->listclass->master->last(bucket->u.list)
						: bucket->u.list->listclass->master->previous(bucket->u.list,nextnode);

		/*
		 *	If we found something, return it
		 */
		if (nextnode != NULL)
        {
            return nextnode;
        }
    }
    return NULL;
}


/*============================================================ HASH_NEXT */
extern
node_t  _LIST_HashNext      (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~
 *  To find the next node, we start at the bucket containing
 *	the current node.  We search for further nodes in attached
 *	to that bucket.  If not we move to the next bucket and
 *	repeat the process.  This carries on until we either
 *	find another node or reach the end of the hash table.
 *-----------------------------------------------------------------------*/
{
	_hashh_t hashlist = (_hashh_t)list;
    _hash_t  bucket;
    _hash_t  lastbucket;
    node_t   nextnode;

	/*
	 *	Check if list set up yet.
	 */
	if (hashlist->buckets == NULL)
	{
		errno = LIST_ERROR_NOTSETUP;
		return NULL;
	}

    for (bucket = _LIST_HashBucket(hashlist, node), nextnode = node,
	     lastbucket = &hashlist->buckets[hashlist->bucketcount] ; 
				bucket < lastbucket; 
						bucket++)
    {
		if (hashlist->clonelist == NULL)
		{
			if (bucket->u.node == NULL)
			{
				continue;
			}
			if (bucket->u.node != node)
			{
				return bucket->u.node;
			}
		}
		if (bucket->u.list == NULL)
		{
			continue;
		}
		/*
		 *	Call first/next as appropriate
		 */
		nextnode = (nextnode == NULL)
						? bucket->u.list->listclass->master->first(bucket->u.list)
						: bucket->u.list->listclass->master->next(bucket->u.list,nextnode);
		/*
		 *	If we found something, return it
		 */
		if (nextnode != NULL)
        {
            return nextnode;
        }
    }
    return NULL;
}


/*=========================================================== TREE_AFTER */
extern
int     _LIST_HashAfter     (list_t list, node_t newnode, node_t oldnode)
/*      ~~~~~~~~~~~~~~~
 *  Add a node after the given node.
 *  This will only make sense on a hash table if both
 *  the nodes are part of the same bucket.  If not it
 *  will be treated as an error.
 *-----------------------------------------------------------------------*/
{
	_hashh_t hashlist = (_hashh_t)list;
    _hash_t  newbucket;

	/*
	 *	Check if list set up yet.
	 */
	if (hashlist->buckets == NULL)
	{
		return LIST_ERROR_NOTSETUP;
	}

    newbucket = _LIST_HashBucket(hashlist, newnode);

	if (hashlist->clonelist == NULL)
	{

		if (oldnode != NULL)
		{
			_hash_t  oldbucket = _LIST_HashBucket(hashlist, oldnode);
			if (sizeof (struct _hash) != newbucket - oldbucket)
			{
				return LIST_ERROR_SEQUENCE;
			}
		}
		/*
		 *	There is no attached list, so result
		 *	must simply be the bucket contents
		 */
		if (newbucket->u.node != NULL)
		{
			return LIST_ERROR_SEQUENCE;
		}
		newbucket->u.node = newnode;
		return LIST_ERROR_NONE;
	}

	/*
	 *	There is a linked list to look at,
	 *	so pass find on to that level.
	 */
	if (HASH_CheckBucket(hashlist, newbucket) != LIST_ERROR_NONE)
	{
		return (errno);
	}
    return newbucket->u.list->listclass->master->after(newbucket->u.list, 
													   newnode,
													   oldnode);
}


/*=========================================================== HASH_ADD */
extern
int     _LIST_HashAdd       (list_t list, node_t node)
/*      ~~~~~~~~~~~~~
 *  Add a node at the correct point.in the hash table.
 *	This invloves finding the correct bucket, and then
 *	adding the node to the data in that bucket.
 *----------------------------------------------------------------------*/
{
	_hashh_t hashlist = (_hashh_t)list;
    _hash_t  bucket;

	/*
	 *	Check if list set up yet.
	 */
	if (hashlist->buckets == NULL)
	{
		return LIST_ERROR_NOTSETUP;
	}

    bucket = _LIST_HashBucket(hashlist, node);

	if (hashlist->clonelist == NULL)
	{
		/*
		 *	There is no attached list, so result
		 *	must simply be the bucket contents
		 */
	    if (bucket->u.node != NULL)
		{
			return LIST_ERROR_NOTEMPTY;
		}
		bucket->u.node = node;
		return (LIST_ERROR_NONE);
	}

	/*
	 *	There is a linked list to look at,
	 *	so pass find on to that level.
	 */

	if (HASH_CheckBucket(hashlist, bucket) != LIST_ERROR_NONE)
	{
		return (errno);
	}
    return LIST_Add ( bucket->u.list,
					  _LIST_USERNODE(hashlist, node));
}


/*========================================================== HASH_REMOVE */
extern
int     _LIST_HashRemove    (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~~~
 *  Remove a node from the list.
 *
 *  This is easy if it is a leaf node, 
 *  but rather more difficult if it is not.
 *-----------------------------------------------------------------------*/
{   
	_hashh_t hashlist = (_hashh_t)list;
    _hash_t  bucket;

	/*
	 *	Check if list set up yet.
	 */
	if (hashlist->buckets == NULL)
	{
		return LIST_ERROR_NOTSETUP;
	}

    bucket = _LIST_HashBucket(hashlist, node);

	if (hashlist->clonelist == NULL)
	{
		/*
		 *	There is no attached list, so result
		 *	must simply be the bucket contents
		 */
		bucket->u.node = NULL;
		return LIST_ERROR_NONE;
	}

	/*
	 *	There is a linked list to look at,
	 *	so pass find on to that level.
	 */
    return bucket->u.list->listclass->master->remove(bucket->u.list,node);
}


/*=========================================================== HASH_FIND */
extern
node_t  _LIST_HashFind (list_t list, int (*findfunc)(node_t, va_list), va_list ap)
/*      ~~~~~~~~~~~~~~
 *  This routine uses the special 'findhash' variant
 *	of the hash algorithm to locate the correct
 *	bucket, and then searches that to find the
 *	wanted node.
 *-----------------------------------------------------------------------*/
{
    _hashh_t hashlist = (_hashh_t)list;
    _hash_t  bucket;

	/*
	 *	Check if list set up yet.
	 */
	if (hashlist->buckets == NULL)
	{
		errno = LIST_ERROR_NOTSETUP;
		return NULL;
	}

	LIBASSERT(hashlist->bucketcount != 0)
    bucket = &hashlist->buckets[hashlist->findhash(ap) % hashlist->bucketcount];

	if (hashlist->clonelist == NULL)
	{
		/*
		 *	There is no attached list, so result
		 *	must simply be the bucket contents
		 */
		return bucket->u.node;
	}

	/*
	 *	There is an attached list.
	 *	However has this bucket got one attached yet?
	 *	If not there is obviously no data in the list.
	 */
	if (bucket->u.list == NULL)
	{
		return NULL;
	}
	/*
	 *	There is a linked list to look at,
	 *	so pass find on to that level.
	 */
    return bucket->u.list->listclass->master->find (bucket->u.list,findfunc,ap);
}
